#include "elslist.h" 
#include "sparsesd.h"
#include "blocksparse.h"
#include <string.h>
#include <stdio.h>
#include <float.h>
#define ZEROTEST(A) ((A) < DBL_EPSILON) /*macro for testing if equal to zero*/
#define REAL_C 1  /* 1 if C is real, 0 if C is imaginary
		   The compiler flag allows user to take advantage of
		   the type of partC, thus using less memory and a little
		   less time */

struct real_block_matrix blockc; 
integer num, na, nb, nc, permuteflag;
doublecomplex *parta;
doublecomplex *partb;
double *partcRealOnly;
integer *partar, *partac, *partbr, *partbc;

/****************************************************************************

  speigprob solves the semi-definite, Hermitian, generalized 
  eigenvalue problem: 
  
                       M.x=lambda.E.x
  
  for a sparse Hermitian matrix M, and positive semidefinite Hermitian
  matrix E.  Currently, E must also be diagonal.
  Calculates a number of extreme eigenvalues using lanczos
  method.
  NOTE: compile with lanso, long, blas, lapack libraries, as well as
        spooles library (however there are bugs in spooles which could
	potentially cause problems??).

  Arguments:
      elslist *els:   contains matrix M, and innerproduct matrix E
      integer maxprs: number of ritz values to be calculated
      integer lanmax: maximum number of lanczos steps to be used
      integer lohi:   flag determining which extreme eigenvalues to calculate
                      = 0:  lowest eigenvalues
		      = 1:  highest eigenvaules
      integer *j:      number of lanczos steps actually taken
      integer *neig:  number of eigenvalues actually calculated
      doublereal *ritz:  sorted extreme eigenvalues calculated
      doublereal *bnd:   hold error bounds
      integer job:    flag determining what to calculate
                      = 0:  compute eigenvalues only;
                      = 1:  compute eigenvalues and vectors w/o constraints
                      = 2:  compute eigenvalues and vectors with constraints
      doublecomplex *vec:  holds eigenvectors if calculated
      integer msglvl: message level for lanczos
      integer *ierr:  error flag
      integer *info:  error flag

*****************************************************************************/
void speigprob(elslist *els, integer maxprs, integer lanmax, integer lohi,
	       integer *j, integer *neig, doublereal *ritz, doublereal *bnd,
	       integer job, doublecomplex *vec, integer msglvl, 
	       integer *ierr, integer *info)
{
  /* declare variables */

  int row, i;

  /* count number of zeros in els->inner */

  num = 0;
  *info = 0;
  for (i = 0; i < els->n; i++) {
    if (els->inner[i] == 0)
      ++num;
    if (els->inner[i] < 0) {
      *info = i;
      return;
    }
  }

  /* count number of elements in three block matrices */

  na=0;
  nb=0;
  nc=0;
  for(row=0; row<els->n; row++)
  for (i = els->j[row]; i < els->j[row+1]; i++) {
    if (els->inner[row] != 0) {
      if (els->inner[els->i[i]] != 0)
	++na;
      else
	++nb;
    }
    else {
      if (els->inner[els->i[i]] != 0)
	++nb;
      else
	++nc;
    }
  }

  /* calculate eigenvalues and eigenvectors */

  speigprobA(els, maxprs, lanmax, lohi, j,
	     neig, ritz, bnd, job, vec, msglvl, ierr);

  return;
}

/*****************************************************************************

  speigproba  is called by speigprob to solve the semi-definite, Hermitian,
  generalized eigenvalue problem for sparse matrices, but only for the
  case where E is diagonal. 
  
  Arguments:
      elslist *els:   contains matrix M, and innerproduct matrix E
      integer maxprs: number of ritz values to be calculated
      integer lanmax: maximum number of lanczos steps to be used
      integer lohi:   flag determining which extreme eigenvalues to calculate
                      = 0:  lowest eigenvalues
		      = 1:  highest eigenvaules
      integer j:      number of lanczos steps actually taken
      integer *neig:  number of eigenvalues actually calculated
      doublereal *ritz:  sorted extreme eigenvalues calculated
      doublereal *bnd:   hold error bounds
      integer job:    flag determining what to calculate
                      = 0:  compute eigenvalues only;
                      = 1:  compute eigenvalues and vectors w/o constraints
                      = 2:  compute eigenvalues and vectors with constraints
      doublecomplex *vec:  holds eigenvectors if calculated
      integer msglvl: message level for lanczos
      integer *ierr:  error flag

*****************************************************************************/
void speigprobA(elslist *els, integer maxprs, integer lanmax, integer lohi,
		integer *j, integer *neig, doublereal *ritz, doublereal *bnd,
		integer job, doublecomplex *vec, integer msglvl,
		integer *ierr)
{
  /* declare variables and initialize values */

  integer order = els->n - num; /* order of reduced matrix */
  int zcount, nzcount, i, k, kount, error, row, ai, ci, ar, cr,
    aknt, bknt, cknt;
  size_t tmplength;

  /*
    Allocate memory
    Make sure els has enough memory to store
    blocks a and b.
  */

  if(els->rc_bytes<(tmplength=(na+nb)*sizeof(*parta)))
    els->c=realloc(els->c,els->rc_bytes=tmplength);
  if(els->c==NULL)goto error;
  if(els->i_bytes<(tmplength=(na+nb)*sizeof(*partac)))
    els->i=realloc(els->i,els->i_bytes=tmplength);
  if(els->i==NULL)goto error;
  if(els->j_bytes<(tmplength=(na+nb)*sizeof(*partar)))
    els->j=realloc(els->j,els->j_bytes=tmplength);
  if(els->j==NULL)goto error;
  int *zeros = malloc(num * sizeof(*zeros));
  int *nonzeros = malloc(order * sizeof(*nonzeros));
  parta = malloc(na * sizeof(*parta));
  partac = malloc(na * sizeof(*partac));
  partar = malloc(na * sizeof(*partar));
  partb = malloc(nb * sizeof(*partb));
  partbc = malloc(nb * sizeof(*partbc));
  partbr = malloc(nb * sizeof(*partbr));
  elslist partc;
  partc.j = malloc((num+1)*sizeof(*partc.j));
  partc.i = malloc(nc*sizeof(*partc.i));
#if REAL_C
  partc.r = malloc(nc * sizeof(*partc.r)); /*real only uses half
						  as much memory*/
  partc.c = NULL;
#else
  partc.c = malloc(nc * sizeof(*partc.c));
  partc.r = NULL;
#endif
  partc.inner=NULL;
  doublecomplex *q=NULL;
  integer *list = malloc(els->n * sizeof(integer));
  ai = ci = kount =  0;
  doublereal kappa = 0;
  permuteflag = 1;
  doublereal condm = 1;
  integer retvec = 0;
  /*  doublereal *transfer;*/
  error = 0;

  /* find zeros and nonzeros of els->inner, and calculate 1/sqrt(els->inner) */

  zcount = 0;
  nzcount = 0;
  for (i = 0; i < els->n; i++) {
    if (ZEROTEST(els->inner[i])) {
      zeros[zcount] = i;
      ++zcount;
    }
    else {
      nonzeros[nzcount] = i;
      els->inner[i] = 1 / sqrt(els->inner[i]);
      ++nzcount;
    }
  }
 
  /* make list of indices for placing values in respective matrices */
  
  for(i = 0; i < els->n; i++) {
    if(ZEROTEST(els->inner[i]))
      list[i] = ci++;
    else
      list[i] = ai++;
  }
  if(els->n != (ai + ci))
    return;
  
  /* make block matrices A, B, and C from large matrix M */
  
  aknt=0; bknt=0; cknt=0;
  for(cr=0, ar=0, row=0; row<els->n; 
      (ZEROTEST(els->inner[row])?++cr:++ar), row++){

    if(ZEROTEST(els->inner[row]))
	partc.j[cr] = cknt;
       
    for(i = els->j[row]; i<els->j[row+1]; i++) {
      
      /* part C */
      if(ZEROTEST(els->inner[row]) && ZEROTEST(els->inner[els->i[i]])) {
#if REAL_C
	partc.r[cknt] = els->c[i].re;
	if(els->c[i].im != 0) {
	  printf("ERROR:  C is IMAGINARY, NOT JUST REAL!!!!!");/*error check*/
	  return;
	}
#else
	partc.c[cknt].re = els->c[i].re;
	partc.c[cknt].im = els->c[i].im;      
#endif
	partc.i[cknt] = list[els->i[i]];
	cknt++;
      }
      /* part B dagger */
      else if(ZEROTEST(els->inner[row]) && !ZEROTEST(els->inner[els->i[i]])) {
	partb[bknt].re = els->c[i].re * els->inner[els->i[i]];
	partb[bknt].im = -els->c[i].im * els->inner[els->i[i]];
	partbr[bknt] = list[els->i[i]];
	partbc[bknt] = cr;
	bknt++;
      }
      /* part B */
      else if(!ZEROTEST(els->inner[row]) && ZEROTEST(els->inner[els->i[i]])) {
	partb[bknt].re = els->c[i].re * els->inner[row];
	partb[bknt].im = els->c[i].im * els->inner[row];
	partbr[bknt] = ar;
	partbc[bknt] = list[els->i[i]];
	bknt++;
      }
      /* part A */
      else {
	parta[aknt].re = els->c[i].re * els->inner[row] * els->inner[els->i[i]];
	parta[aknt].im = els->c[i].im * els->inner[row] * els->inner[els->i[i]];
	partar[aknt] = ar;
	partac[aknt] = list[els->i[i]];
	aknt++;
      }
    }
  }
  partc.n=num;
  partc.j[num]=cknt;  
  
  /* free extra memory and move arrays to elslist */

  free(list);
  memcpy(els->c,parta,na*sizeof(*parta));
  free(parta); parta=(doublecomplex *) els->c;
  memcpy(els->i,partar,na*sizeof(*partar));
  free(partar); partar=(integer *) els->i;
  memcpy(els->j,partac,na*sizeof(*partac));
  free(partac); partac=(integer *) els->j;

  memcpy(parta+na,partb,nb*sizeof(*partb));
  free(partb); partb=parta+na;
  memcpy(partar+na,partbr,nb*sizeof(*partbr));
  free(partbr); partbr=partar+na;
  memcpy(partac+na,partbc,nb*sizeof(*partbc));
  free(partbc); partbc=partac+na;

#if 0   /* print out sparse matrix */
  {
    int ii,jj;
    printf("sparse matrix {colp,i,r}:");
    for(ii=0; ii<partc.j[partc.n]; ii++)
      printf("%s {%i,%i,%g}",ii?",":"",ii,partc.i[ii],partc.r[ii]);
    printf("\nj:  ");
    for(jj=0; jj<=partc.n; jj++)
      printf(" %s%i\n",jj?",":"",partc.j[jj]);
    printf("\n");
  }
#endif

#if REAL_C
  sparse_to_real_block_matrix(&partc,&blockc);
#else
  fprintf(stderr," not done yet\n");
#endif
  free_els(&partc);

  /*******************************************************

    call Lanczos to find eigenvalues and vectors 
  
  *********************************************************/

  q = malloc(tmplength=order * lanmax * sizeof(*q));
  if(q==NULL){
    fprintf(stderr,"  not allocated\n");
    return;
  }

  for(i = 0; i < order * lanmax; i++) {
    q[i].re = 0;
    q[i].im = 0;
  }
  /* note:  NEED TO CHANGE TO FORTRAN INDICES IN partar, partac,
            partbr, partbc FOR USE IN ZOP IN ZLANDR3!!!!*/
  for(i = 0; i < nb; i++) {
    partbr[i]++;
    partbc[i]++;
  }
  for(i = 0; i < na; i++) {
    partar[i]++;
    partac[i]++;
  }

  /* set return vector flag based on job # */
  if(job > 0) retvec = 1;
  ZLANDR3(&order, &lanmax, &maxprs, &lohi, &condm, &kappa, j, neig, ritz,
 	  bnd, &retvec, vec, q, ierr, &msglvl);

  /* calculate correct eigenvectors for job # */
  /* The eigenvectors are stored in reduced form */

  if(job == 2) {
    /* allocate memory for temporary workspace vectors */
    doublecomplex *xtmp = malloc(order * sizeof(*xtmp));
    doublecomplex *ytmp = malloc(num * sizeof(*ytmp));

    for(i=maxprs-1; i>=0; i--) {
      /* store X in of xtmp */
      for(k = 0; k < order; k++) {
	xtmp[k].re = vec[(order * i) + k].re;
	xtmp[k].im = vec[(order * i) + k].im;
	/* resize eigenvectors X */
	vec[(order * i) + k].re = 
	  vec[(order * i) + k].re /  sqrt(els->inner[nonzeros[k]]);
 	vec[(order * i) + k].im = 
	  vec[(order * i) + k].im /  sqrt(els->inner[nonzeros[k]]);
      }
      /*===calculate eigenvectors Y in ytmp====*/
      char trans = 'C';
      ZSMVM(&trans, &num, partb, partbr, partbc, &nb, vec + (order * i),
	    &order, ytmp, 1);
      
      /* create and fill right hand side dense matrix */
      
      /* sort X and Y back into M in the correct n */
      for(k = 0; k < num; k++) {
	vec[els->n * i + zeros[k]].re = -ytmp[k].re;
	vec[els->n * i + zeros[k]].im = -ytmp[k].im;
      }
      for(k = 0; k < order; k++) {
	vec[els->n * i + nonzeros[k]].re = xtmp[k].re;
	vec[els->n * i + nonzeros[k]].im = xtmp[k].im;
      }
    }
    free(xtmp);
    free(ytmp);
  }
  goto free;

 error:
  fprintf(stderr,__FILE__":  Memory allocation error \n");
  

  /* free extra memory */

 free:
  free(q);
  free_real_block_matrix(&blockc);
  free(zeros);
  free(nonzeros);

  return;
}

/***************************************************************************

   ZOP is the linear Hermitian operator supplied to the complex lanczos
   routine, zlandr3; it calculates P=(A-B.Cinverse.Bdagger).Q

   Arguments:
       integer *n:        length of P, Q, and R
       doublecomplex *P:  unused in this particular operator
       doublecomplex *Q:  lanczos vector being acted on by operator
       doublecomplex *R:  result of operator acting on lanczos vector

****************************************************************************/
extern fsub ZOP(integer *n, doublecomplex *P, doublecomplex *Q,
	 	doublecomplex *R)
{
  /* declare variables */

  int i;
  char trans = 'C';
  integer incr = 1;
  /*  doublereal *transfer; */
  doublecomplex konst = {-1, 0};
  doublecomplex *tmp = malloc((*n + num) * sizeof(*tmp));
#if REAL_C
  doublereal *realpart = malloc(num*sizeof(*realpart));
#endif
  /* calculate tmp = Bdagger.Q */

  ZSMVM(&trans, &num, partb, partbr, partbc, &nb, Q, n, tmp, 1);
  
#if REAL_C
  for(i=0; i<num; i++)realpart[i]=tmp[i].re;
  solve_real_block(&blockc,num,realpart);
  for(i=0; i<num; i++)tmp[i].re=realpart[i];
  for(i=0; i<num; i++)realpart[i]=tmp[i].im;
  solve_real_block(&blockc,num,realpart);
  for(i=0; i<num; i++)tmp[i].im=realpart[i];
#else
  fprintf(stderr,"Code not written\n");
#endif

  /* calculate R = B.tmp */

  trans = 'N';
  ZSMVM(&trans, n, partb, partbr, partbc, &nb, tmp, &num, R, 1);

  /* calculate tmp = A.Q */

  ZSHMVM(n, parta, partar, partac, &na, Q, n, tmp);

  /* calculate R = tmp - R */

  ZAXPY(n, &konst, tmp, &incr, R, &incr); 


  /* note: zaxpy is R = R - tmp, so must multiply by -1 */
  for(i = 0; i < *n; i++) {
    R[i].re = -R[i].re;
    R[i].im = -R[i].im;
  }

  /* free extra memory */

  free(tmp);
#if REAL_C
  free(realpart);
#endif
  return SUBRETURN;
}

/*************************************************************************

   ZOPM calculates the inner product in the lanczos routine, zlandr3, for
   double precision complex vectors; copies input vector to output vector

   Arguments:
       integer *n:        length of x and y
       doublecomplex *x:  input vector
       doublecomplex *y:  output vector

**************************************************************************/
extern fsub ZOPM(integer *n, doublecomplex *x, doublecomplex *y)
{
  memcpy(y, x, (*n) * sizeof(doublecomplex));
  return SUBRETURN;
}

/**************************************************************************

   junk function so code will compile

**************************************************************************/
extern fsub OP(integer *n, doublereal *P, doublereal *Q,
		doublereal *R)
{
  fprintf(stderr,"Real OP not written\n");
  exit(22);
  return SUBRETURN;
}

/**************************************************************************

   junk fuction so code will compile 

**************************************************************************/
extern fsub OPM(integer *n, doublereal *x, doublereal *y)
{
  memcpy(y, x, (*n) * sizeof(doublereal));
  return SUBRETURN;
}

